"""generate openapi docs."""
from pkg_resources import get_distribution
from honeybee_schema._openapi import get_openapi, get_model_mapper
from honeybee_schema.model import Model
from honeybee_schema.energy.simulation import SimulationParameter

import json
import argparse

parser = argparse.ArgumentParser(description='Generate OpenAPI JSON schemas')

parser.add_argument('--version',
                    help='Set the version of the new OpenAPI Schema')

args = parser.parse_args()

VERSION = None

if args.version:
    VERSION = args.version.replace('v', '')
else:
    try:
        VERSION = '.'.join(get_distribution('honeybee_schema').version.split('.')[:3]),
    except:
        pass

info = {
        "description": "",
        "version": VERSION,
        "title": "",
        "contact": {
            "name": "Ladybug Tools",
            "email": "info@ladybug.tools",
            "url": "https://github.com/ladybug-tools/honeybee-schema"
        },
        "x-logo": {
            "url": "https://www.ladybug.tools/assets/img/honeybee-large.png",
            "altText": "Honeybee logo"
        },
        "license": {
            "name": "BSD",
            "url": "https://github.com/ladybug-tools-in2/honeybee-schema/blob/master/LICENSE"
        }
}


# generate Model open api schema
print('Generating Model documentation...')

external_docs = {
    "description": "OpenAPI Specification with Inheritance",
    "url": "./model_inheritance.json"
}

openapi = get_openapi(
    [Model],
    title='Honeybee Model Schema',
    description='This is the documentation for Honeybee model schema.',
    version=VERSION, info=info,
    external_docs=external_docs)
with open('./docs/model.json', 'w') as out_file:
    json.dump(openapi, out_file, indent=2)

# with inheritance
openapi = get_openapi(
    [Model],
    title='Honeybee Model Schema',
    description='This is the documentation for Honeybee model schema.',
    version=VERSION, info=info,
    inheritance=True,
    external_docs=external_docs
)

with open('./docs/model_inheritance.json', 'w') as out_file:
    json.dump(openapi, out_file, indent=2)

# add the mapper file
mapper = get_model_mapper(Model)
module_mapper = {k: c.__module__ for k, c in mapper.items()}
with open('./docs/model_mapper.json', 'w') as out_file:
    json.dump(module_mapper, out_file, indent=2)

# generate SimulationParameter open api schema
print('Generating Energy Simulation Parameter documentation...')

external_docs = {
    "description": "OpenAPI Specification with Inheritance",
    "url": "./simulation-parameter_inheritance.json"
}

openapi = get_openapi(
    [SimulationParameter],
    title='Honeybee Energy Simulation Parameter Schema',
    description='This is the documentation for Honeybee energy simulation parameter schema.',
    version=VERSION, info=info,
    external_docs=external_docs)
with open('./docs/simulation-parameter.json', 'w') as out_file:
    json.dump(openapi, out_file, indent=2)

openapi = get_openapi(
    [SimulationParameter],
    title='Honeybee Energy Simulation Parameter Schema',
    description='This is the documentation for Honeybee energy simulation parameter schema.',
    version=VERSION, inheritance=True, info=info,
    external_docs=external_docs
)
with open('./docs/simulation-parameter_inheritance.json', 'w') as out_file:
    json.dump(openapi, out_file, indent=2)

# add the mapper file
mapper = get_model_mapper(SimulationParameter)
module_mapper = {k: c.__module__ for k, c in mapper.items()}
with open('./docs/simulation-parameter_mapper.json', 'w') as out_file:
    json.dump(module_mapper, out_file, indent=2)



from pydantic.utils import get_model
from pydantic.schema import schema, get_flat_models_from_model, get_model_name_map
from typing import Dict

# base open api dictionary for all schemas
_base_open_api = {
    "openapi": "3.0.2",
    "servers": [],
    "info": {},
    "externalDocs": {},
    "tags": [],
    "x-tagGroups": [
        {
            "name": "Models",
            "tags": []
        }
    ],
    "paths": {},
    "components": {"schemas": {}}
}


def get_openapi(
    base_object,
    title: str = None,
    version: str = None,
    openapi_version: str = "3.0.2",
    description: str = None,
    info: dict = None,
    external_docs: dict = None,
    inheritance: bool = False
        ) -> Dict:
    """Return Honeybee Model Schema as an openapi compatible dictionary."""
    open_api = dict(_base_open_api)

    open_api['openapi'] = openapi_version

    if title:
        open_api['info']['title'] = title

    if not version:
        raise ValueError(
            'Schema version must be specified as argument or from distribution metadata'
        )

    if info:
        open_api['info'] = info

    if version:
        open_api['info']['version'] = version

    if description:
        open_api['info']['description'] = description

    if external_docs:
        open_api['externalDocs'] = external_docs

    if not inheritance:
        schemas = schema(base_object, ref_prefix='#/components/schemas/')['definitions']
    else:
        schemas = get_schemas_inheritance(base_object)

    # goes to tags
    tags = []
    # goes to x-tagGroups['tags']
    tag_names = []

    schema_names = list(schemas.keys())
    schema_names.sort()

    for name in schema_names:
        model_name = '%s_model' % name.lower()
        tag_names.append(model_name)
        tag = {
            'name': model_name,
            'x-displayName': name,
            'description':
                '<SchemaDefinition schemaRef=\"#/components/schemas/%s\" />\n' % name
        }
        tags.append(tag)

        # sort properties order: put required parameters at begining of the list
        s = schemas[name]

        if 'properties' in s:
            properties = s['properties']
        else:
            properties = s['allOf'][1]['properties']

        # make all types readOnly
        try:
            properties['type']['readOnly'] = True
        except KeyError:
            # no type has been set in properties for this object
            typ = {
                'title': 'Type', 'default': f'{name}', 'type': 'string',
                'pattern': f'^{name}$', 'readOnly': True,
            }
            properties['type'] = typ
        # add format to numbers and integers
        # this is helpful for C# generators
        for prop in properties:
            try:
                properties[prop] = set_format(properties[prop])
            except KeyError:
                # referenced object
                if 'anyOf' in properties[prop]:
                    new_any_of = []
                    for item in properties[prop]['anyOf']:
                        new_any_of.append(set_format(item))
                    properties[prop]['anyOf'] = new_any_of
                else:
                    continue

        # sort fields to keep required ones on top
        if 'required' in s:
            required = s['required']
        elif 'allOf' in s:
            try:
                required = s['allOf'][1]['required']
            except KeyError:
                # no required field
                continue
        else:
            continue

        sorted_props = {}
        optional = {}
        for prop, value in properties.items():
            if prop in required:
                sorted_props[prop] = value
            else:
                optional[prop] = value

        sorted_props.update(optional)

        try:
            s['properties'] = sorted_props
        except KeyError:
            s['allOf'][1]['properties'] = sorted_props

    tag_names.sort()
    open_api['tags'] = tags
    open_api['x-tagGroups'][0]['tags'] = tag_names

    open_api['components']['schemas'] = schemas

    return open_api


def set_format(p):
    """Set format for a property."""
    if '$ref' in p:
        return p
    elif p['type'] == 'number' and 'format' not in p:
        p['format'] = 'double'
    elif p['type'] == 'integer' and 'format' not in p:
        p['format'] = 'int32'
    elif p['type'] == 'array':
        p['items'] = set_format(p['items'])
    return p


def get_model_mapper(model, stoppage=None, full=True):
    """Get a dictionary of name: class for all the objects in model."""
    model = get_model(model)
    flat_models = get_flat_models_from_model(model)

    # this is the list of all the referenced objects
    model_name_map = get_model_name_map(flat_models)
    # flip the dictionary so I can access each class by name
    model_name_map = {v: k for k, v in model_name_map.items()}

    if full:
        if not stoppage:
            stoppage = set(['NoExtraBaseModel', 'ModelMetaclass', 'BaseModel', 'object'])
        # Pydantic does not necessarily add all the baseclasses to the OpenAPI
        # documentation. We check all of them and them to the list if they are not
        # already added
        models = list(model_name_map.values())
        for model in models:
            for cls in type.mro(model):
                if cls.__name__ in stoppage:
                    break
                if cls.__name__ not in model_name_map:
                    model_name_map[cls.__name__] = cls

    return model_name_map


def get_schemas_inheritance(model_cls):
    """This method modifies the default OpenAPI from Pydantic.

    It adds referenced values to subclasses using allOf field as explained in this post:
    https://swagger.io/docs/specification/data-models/inheritance-and-polymorphism
    """
    # list of top level class names that we should stop at
    stoppage = set(['NoExtraBaseModel', 'ModelMetaclass', 'BaseModel', 'object'])

    model_name_map = get_model_mapper(model_cls[0], stoppage, full=True)

    # get the standard OpenAPI schema for Pydantic for all the new objects
    ref_prefix = '#/components/schemas/'
    schemas = \
        schema(model_name_map.values(), ref_prefix=ref_prefix)['definitions']

    # collect updated objects
    updated_schemas = {}

    # iterate through all the data models
    # find the ones which are subclassed and updated them based on the properties of
    # baseclasses.
    for name in schemas.keys():
        # find the class from class name
        main_cls = model_name_map[name]
        top_classes = []
        for cls in type.mro(main_cls):
            if cls.__name__ in stoppage:
                break
            top_classes.append(cls)
        if len(top_classes) < 2:
            # this class is not a subclass
            print(f'\n{name} is not a subclass.')
            continue
        # set object inheritance
        updated_schema = set_inheritance(name, top_classes, schemas)
        updated_schemas[name] = updated_schema

    # replace updated schemas in original schema
    for name, value in updated_schemas.items():
        schemas[name] = value

    return schemas


def set_inheritance(name, top_classes, schemas):
    """Set inheritance for object with a certain name."""
    # this is the list of special keys that we copy in manually
    copied_keys = set(['type', 'properties', 'required'])
    # remove the class itself
    print(f'\nProcessing {name}')
    top_classes = top_classes[1:]
    top_class = top_classes[0]
    tree = ['....' * (i + 1) + c.__name__ for i, c in enumerate(top_classes)]
    print('\n'.join(tree))

    # collect required and properties from top classes so we don't end up with
    # duplicate values in the schema for the subclass
    top_classes_required = []
    top_classes_prop = {}

    for t in top_classes:
        if 'required' in schemas[t.__name__]:
            tc_required = schemas[t.__name__]['required']
        else:
            continue

        for r in tc_required:
            top_classes_required.append(r)

    for t in top_classes:
        tc_prop = schemas[t.__name__]['properties']

        for pn, dt in tc_prop.items():
            # collect type to find the cases when the property has the same name
            # but is defined with a different type
            if 'type' in dt:
                top_classes_prop[pn] = dt['type']
            else:
                top_classes_prop[pn] = '###'  # no type means use of oneOf or allOf

    # create a new schema for this object based on the top level class
    data = {
        'allOf': [
            {
                '$ref': f'#/components/schemas/{top_class.__name__}'
            },
            {
                'type': 'object',
                'required': [],
                'properties': {}
            }
        ]
    }

    data_copy = dict(data)
    # the immediate top class openapi schema
    object_dict = schemas[name]

    if not top_classes_required and 'required' in object_dict:
        # no required in top level class
        for r in object_dict['required']:
            data_copy['allOf'][1]['required'].append(r)
    elif 'required' in object_dict and top_classes_required:
        for r in object_dict['required']:
            if r not in top_classes_required:
                data_copy['allOf'][1]['required'].append(r)

    # no required fields
    if len(data_copy['allOf'][1]['required']) == 0:
        del(data_copy['allOf'][1]['required'])

    properties = object_dict['properties']

    for prop in properties:
        if prop not in top_classes_prop:
            # new field. add it to the properties
            print(f'Extending: {prop}')
            data_copy['allOf'][1]['properties'][prop] = properties[prop]
        elif 'type' not in properties[prop] and \
                ('allOf' in properties[prop] or 'anyOf' in properties[prop]):
            # same name diffrent types
            print(f'Found a field with the same name: {prop}.')
            if len(top_classes) > 1:
                print(f'Trying {name} against {top_classes[1].__name__}.')
                return set_inheritance(name, top_classes, schemas)
            else:
                return schemas[name]

    try:
        data_copy['allOf'][1]['properties']['type'] = properties['type']
    except KeyError:
        print(f'Found object with no type:{name}')

    # add other items in addition to copied_keys
    for key, value in schemas[name].items():
        if key in copied_keys:
            continue
        data_copy[key] = value

    return data_copy



from honeybee_schema.energy.material import EnergyMaterial, EnergyMaterialNoMass, \
    EnergyWindowMaterialGas, EnergyWindowMaterialGasMixture, \
    EnergyWindowMaterialGasCustom, EnergyWindowMaterialBlind, \
    EnergyWindowMaterialGlazing, EnergyWindowMaterialShade, \
    EnergyWindowMaterialSimpleGlazSys
from copy import copy
from pydantic import ValidationError
import pytest
import os
import json

# target folder where all of the samples live
root = os.path.dirname(os.path.dirname(__file__))
target_folder = os.path.join(root, 'samples', 'material')


def test_material_opaque_gypsum():
    file_path = os.path.join(target_folder, 'material_opaque_gypsum.json')
    EnergyMaterial.parse_file(file_path)


def test_material_opaque_insulation():
    file_path = os.path.join(target_folder, 'material_opaque_insulation.json')
    EnergyMaterial.parse_file(file_path)


def test_material_opaque_concrete():
    file_path = os.path.join(target_folder, 'material_opaque_concrete.json')
    EnergyMaterial.parse_file(file_path)


def test_material_opaque_brick():
    file_path = os.path.join(target_folder, 'material_opaque_brick.json')
    EnergyMaterial.parse_file(file_path)


def test_material_opaque_wall_gap():
    file_path = os.path.join(target_folder, 'material_opaque_wall_gap.json')
    EnergyMaterialNoMass.parse_file(file_path)


def material_window_glazing_clear():
    file_path = os.path.join(target_folder, 'material_window_glazing_clear.json')
    EnergyWindowMaterialGlazing.parse_file(file_path)


def material_window_glazing_lowe():
    file_path = os.path.join(target_folder, 'material_window_glazing_lowe.json')
    EnergyWindowMaterialGlazing.parse_file(file_path)


def test_material_window_blind():
    file_path = os.path.join(target_folder, 'material_window_blind.json')
    EnergyWindowMaterialBlind.parse_file(file_path)


def test_material_window_gas():
    file_path = os.path.join(target_folder, 'material_window_gas.json')
    EnergyWindowMaterialGas.parse_file(file_path)


def test_material_window_gas_mixture():
    file_path = os.path.join(target_folder, 'material_window_gas_mixture.json')
    EnergyWindowMaterialGasMixture.parse_file(file_path)


def test_material_window_gas_custom():
    file_path = os.path.join(target_folder, 'material_window_gas_custom.json')
    EnergyWindowMaterialGasCustom.parse_file(file_path)


def test_material_window_glazing_system():
    file_path = os.path.join(target_folder, 'material_window_glazing_system.json')
    EnergyWindowMaterialSimpleGlazSys.parse_file(file_path)


def test_material_window_shade():
    file_path = os.path.join(target_folder, 'material_window_shade.json')
    EnergyWindowMaterialShade.parse_file(file_path)


def test_material_wrong():
    file_path = os.path.join(target_folder, 'material_opaque_gypsum.json')
    with open(file_path) as json_file:
        material_gypsum = json.load(json_file)
    wrong_name = copy(material_gypsum)
    wrong_name['name'] = ''
    with pytest.raises(ValidationError):
        EnergyMaterial.parse_obj(wrong_name)
    wrong_thickness = copy(material_gypsum)
    wrong_thickness['thickness'] = 5
    with pytest.raises(ValidationError):
        EnergyMaterial.parse_obj(wrong_thickness)
    wrong_specificheat = copy(material_gypsum)
    wrong_specificheat['specific_heat'] = 0
    with pytest.raises(ValidationError):
        EnergyMaterial.parse_obj(wrong_specificheat)


def test_materialnomass_wrong():
    file_path = os.path.join(target_folder, 'material_opaque_wall_gap.json')
    with open(file_path) as json_file:
        material_no_mass = json.load(json_file)
    wrong_r_value = copy(material_no_mass)
    wrong_r_value['r_value'] = 0
    with pytest.raises(ValidationError):
        EnergyMaterialNoMass.parse_obj(wrong_r_value)
    wrong_solar_absorptance = copy(material_no_mass)
    wrong_solar_absorptance['solar_absorptance'] = 2
    with pytest.raises(ValidationError):
        EnergyMaterialNoMass.parse_obj(wrong_solar_absorptance)


def test_window_simpleglaz_wrong():
    file_path = os.path.join(target_folder, 'material_window_glazing_system.json')
    with open(file_path) as json_file:
        window_simpleglazing = json.load(json_file)
    wrong_values = copy(window_simpleglazing)
    wrong_values['u_factor'] = 6
    with pytest.raises(ValidationError):
        EnergyWindowMaterialSimpleGlazSys.parse_obj(wrong_values)
    wrong_values['SHGC'] = 2
    with pytest.raises(ValidationError):
        EnergyWindowMaterialSimpleGlazSys.parse_obj(wrong_values)


def test_windowshade_wrong():
    file_path = os.path.join(target_folder, 'material_window_shade.json')
    with open(file_path) as json_file:
        window_shade = json.load(json_file)
    wrong_type = copy(window_shade)
    wrong_type['type'] = 'EnergyWindowMaterial'
    with pytest.raises(ValidationError):
        EnergyWindowMaterialShade.parse_obj(wrong_type)
    wrong_shadedistance = copy(window_shade)
    wrong_shadedistance['distance_to_glass'] = 0
    with pytest.raises(ValidationError):
        EnergyWindowMaterialShade.parse_obj(wrong_shadedistance)
    wrong_airflow = copy(window_shade)
    wrong_airflow['airflow_permeability'] = 1
    with pytest.raises(ValidationError):
        EnergyWindowMaterialShade.parse_obj(wrong_airflow)



# coding=utf-8
from honeybee_energy.hvac.idealair import IdealAirSystem
from honeybee_energy.schedule.day import ScheduleDay
from honeybee_energy.schedule.ruleset import ScheduleRuleset
import honeybee_energy.lib.scheduletypelimits as schedule_types

from ladybug.dt import Time

import os
import json


def ideal_air_default(directory):
    ideal_air = IdealAirSystem('Default HVAC System')
    dest_file = os.path.join(directory, 'ideal_air_default.json')
    with open(dest_file, 'w') as fp:
        json.dump(ideal_air.to_dict(abridged=True), fp, indent=4)


def ideal_air_detailed(directory):
    ideal_air = IdealAirSystem('Passive House HVAC System')
    
    ideal_air.economizer_type = 'DifferentialEnthalpy'
    ideal_air.demand_controlled_ventilation = True
    ideal_air.sensible_heat_recovery = 0.75
    ideal_air.latent_heat_recovery = 0.6
    ideal_air.heating_air_temperature = 40
    ideal_air.cooling_air_temperature = 15
    ideal_air.heating_limit = 2000
    ideal_air.cooling_limit = 3500
    sch_day = ScheduleDay('Day Control', [0, 1, 0], [Time(0, 0), Time(8, 0), Time(22, 0)])
    schedule = ScheduleRuleset('HVAC Control', sch_day, None, schedule_types.on_off)
    ideal_air.heating_availability = schedule
    ideal_air.cooling_availability = schedule

    dest_file = os.path.join(directory, 'ideal_air_detailed.json')
    with open(dest_file, 'w') as fp:
        json.dump(ideal_air.to_dict(abridged=True), fp, indent=4)


# run all functions within the file
master_dir = os.path.split(os.path.dirname(__file__))[0]
sample_directory = os.path.join(master_dir, 'samples', 'hvac')

ideal_air_default(sample_directory)
ideal_air_detailed(sample_directory)




"""Construction Schema"""
from pydantic import Field, constr
from typing import List, Union

from ._base import IDdEnergyBaseModel
from .material import EnergyMaterial, EnergyMaterialNoMass, \
    EnergyWindowMaterialGas, EnergyWindowMaterialGasCustom, \
    EnergyWindowMaterialGasMixture, EnergyWindowMaterialSimpleGlazSys, \
    EnergyWindowMaterialBlind, EnergyWindowMaterialGlazing, EnergyWindowMaterialShade
from .schedule import ScheduleRuleset, ScheduleFixedInterval


class WindowConstructionAbridged(IDdEnergyBaseModel):
    """Construction for window objects (Aperture, Door)."""

    type: constr(regex='^WindowConstructionAbridged$') = 'WindowConstructionAbridged'

    layers: List[constr(min_length=1, max_length=100)] = Field(
        ...,
        description='List of strings for material identifiers. The order of the '
            'materials is from exterior to interior.',
        min_items=1,
        max_items=8
    )


class WindowConstruction(WindowConstructionAbridged):
    """Construction for window objects (Aperture, Door)."""

    type: constr(regex='^WindowConstruction$') = 'WindowConstruction'

    materials: List[
        Union[
            EnergyWindowMaterialGas, EnergyWindowMaterialGasCustom, EnergyWindowMaterialGasMixture,
            EnergyWindowMaterialSimpleGlazSys, EnergyWindowMaterialBlind,
            EnergyWindowMaterialGlazing, EnergyWindowMaterialShade
        ]
    ] = Field(
        ...,
        description='List of materials. The order of the materials is from outside '
            'to inside.',
        min_items=1,
        max_items=8
    )


class OpaqueConstructionAbridged(IDdEnergyBaseModel):
    """Construction for opaque objects (Face, Shade, Door)."""

    type: constr(regex='^OpaqueConstructionAbridged$') = 'OpaqueConstructionAbridged'

    layers: List[constr(min_length=1, max_length=100)] = Field(
        ...,
        description='List of strings for material identifiers. The order of the materials '
            'is from exterior to interior.',
        min_items=1,
        max_items=10
    )


class OpaqueConstruction(OpaqueConstructionAbridged):
    """Construction for opaque objects (Face, Shade, Door)."""

    type: constr(regex='^OpaqueConstruction$') = 'OpaqueConstruction'

    materials: List[Union[EnergyMaterial, EnergyMaterialNoMass]] = Field(
        ...,
        description='List of materials. The order of the materials is from outside to'
        ' inside.',
        min_items=1,
        max_items=10
    )


class ShadeConstruction(IDdEnergyBaseModel):
    """Construction for Shade objects."""

    type: constr(regex='^ShadeConstruction$') = 'ShadeConstruction'

    solar_reflectance: float = Field(
        0.2,
        ge=0,
        le=1,
        description=' A number for the solar reflectance of the construction.'
    )

    visible_reflectance: float = Field(
        0.2,
        ge=0,
        le=1,
        description=' A number for the visible reflectance of the construction.'
    )

    is_specular: bool = Field(
        default=False,
        description='Boolean to note whether the reflection off the shade is diffuse '
            '(False) or specular (True). Set to True if the construction is '
            'representing a glass facade or a mirror material.'
    )


class AirBoundaryConstructionAbridged(IDdEnergyBaseModel):
    """Construction for Air Boundary objects."""

    type: constr(regex='^AirBoundaryConstructionAbridged$') = \
        'AirBoundaryConstructionAbridged'

    air_mixing_per_area: float = Field(
        0.1,
        ge=0,
        description='A positive number for the amount of air mixing between Rooms '
            'across the air boundary surface [m3/s-m2]. Default: 0.1 corresponds '
            'to average indoor air speeds of 0.1 m/s (roughly 20 fpm), which is '
            'typical of what would be induced by a HVAC system.'
    )

    air_mixing_schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Identifier of a fractional schedule for the air mixing schedule '
            'across the construction.'
    )


class AirBoundaryConstruction(AirBoundaryConstructionAbridged):
    """Construction for Air Boundary objects."""

    type: constr(regex='^AirBoundaryConstruction$') = 'AirBoundaryConstruction'

    air_mixing_schedule: Union[ScheduleRuleset, ScheduleFixedInterval] = Field(
        ...,
        description='A fractional schedule as a ScheduleRuleset or '
            'ScheduleFixedInterval for the air mixing schedule across '
            'the construction.'
    )

    class Config:
        @staticmethod
        def schema_extra(schema, model):
            schema['properties']['air_mixing_schedule']['anyOf'] = [
                    {"$ref": "#/components/schemas/ScheduleRuleset"},
                    {"$ref": "#/components/schemas/ScheduleFixedInterval"}
                ]


if __name__ == '__main__':
    print(WindowConstructionAbridged.schema_json(indent=2))
    print(WindowConstruction.schema_json(indent=2))
    print(OpaqueConstructionAbridged.schema_json(indent=2))
    print(OpaqueConstruction.schema_json(indent=2))
    print(ShadeConstruction.schema_json(indent=2))
    print(AirBoundaryConstruction.schema_json(indent=2))


"""Simulation Parameter Schema"""
from pydantic import Field, constr, validator
from typing import Union, List
from enum import Enum
import datetime

from .._base import NoExtraBaseModel


class DryBulbCondition(NoExtraBaseModel):
    """Used to specify dry bulb conditions on a design day."""

    type: constr(regex='^DryBulbCondition$') = 'DryBulbCondition'

    dry_bulb_max: float = Field(
        ...,
        ge=-90,
        le=70,
        description='The maximum dry bulb temperature on the design day [C].'
    )

    dry_bulb_range: float = Field(
        ...,
        ge=0,
        description='The difference between min and max temperatures on the' 
            'design day [C].'
    )


class HumidityTypes(str, Enum):
    wetbulb = 'Wetbulb'
    dewpoint = 'Dewpoint'
    humidity_ratio = 'HumidityRatio'
    enthalpy = 'Enthalpy'


class HumidityCondition(NoExtraBaseModel):
    """Used to specify humidity conditions on a design day."""

    type: constr(regex='^HumidityCondition$') = 'HumidityCondition'

    humidity_type: HumidityTypes

    humidity_value: float = Field(
        ...,
        description='The value correcponding to the humidity_type.'
    )

    barometric_pressure: float = Field(
        101325,
        ge=31000,
        le=120000,
        description='Barometric air pressure on the design day [Pa].'
    )

    rain: bool = Field(
        default=False,
        description='Boolean to indicate rain on the design day.'
    )

    snow_on_ground: bool = Field(
        default=False,
        description='Boolean to indicate snow on the ground during the design day.'
    )


class WindCondition(NoExtraBaseModel):
    """Used to specify wind conditions on a design day."""

    type: constr(regex='^WindCondition$') = 'WindCondition'

    wind_speed: float = Field(
        ...,
        ge=0,
        le=40,
        description='Wind speed on the design day [m/s].'
    )

    wind_direction: float = Field(
        0,
        ge=0,
        le=360,
        description='Wind direction on the design day [degrees].'
    )


class _SkyCondition(NoExtraBaseModel):
    """Used to specify sky conditions on a design day."""
    
    date: List[int] = Field(
        ...,
        min_items=2,
        max_items=3,
        description='A list of two integers for [month, day], representing the date '
            'for the day of the year on which the design day occurs.'
            'A third integer may be added to denote whether the date should be '
            're-serialized for a leap year (it should be a 1 in this case).'
    )

    @validator('date')
    def check_date(cls, v):
        "Ensure valid date."
        if len(v) == 3 and v[2]:
            try:
                datetime.date(2016, v[0] , v[1])
            except ValueError:
                raise ValueError('{}/{} is not a valid date.'.format(v[0], v[1]))
        else:
            try:
                datetime.date(2017, v[0] , v[1])
            except ValueError:
                raise ValueError('{}/{} is not a valid date.'.format(v[0], v[1]))
        return v
    
    daylight_savings: bool = Field(
        default=False,
        description='Boolean to indicate whether daylight savings time is active '
            'on the design day.'
    )


class ASHRAEClearSky(_SkyCondition):
    """Used to specify sky conditions on a design day."""

    type: constr(regex='^ASHRAEClearSky$') = 'ASHRAEClearSky'

    clearness: float = Field(
        ...,
        ge=0,
        le=1.2,
        description='Value between 0 and 1.2 that will get multiplied by the '
            'irradinace to correct for factors like elevation above sea level.'
    )


class ASHRAETau(_SkyCondition):
    """Used to specify sky conditions on a design day."""

    type: constr(regex='^ASHRAETau$') = 'ASHRAETau'

    tau_b: float = Field(
        ...,
        ge=0,
        le=1.2,
        description='Value for the beam optical depth. Typically found in .stat files.'
    )

    tau_d: float = Field(
        ...,
        ge=0,
        le=3,
        description='Value for the diffuse optical depth. Typically found in .stat files.'
    )


class DesignDayTypes(str, Enum):
    summer_design_day = 'SummerDesignDay'
    winter_design_day = 'WinterDesignDay'
    sunday = 'Sunday'
    monday = 'Monday'
    tuesday = 'Tuesday'
    wednesday = 'Wednesday'
    thursday = 'Thursday'
    friday = 'Friday'
    holiday = 'Holiday'
    custom_day1 = 'CustomDay1'
    custom_day2 = 'CustomDay2'


class DesignDay(NoExtraBaseModel):
    """An object representing design day conditions."""

    type: constr(regex='^DesignDay$') = 'DesignDay'

    name: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='Text string for a unique design day name. This name remains '
            'constant as the object is mutated, copied, and serialized to different '
            'formats (eg. dict, idf, osm). It is also used to reference the object '
            'within SimulationParameters. It must be < 100 characters, use only '
            'ASCII characters and exclude (, ; ! \\n \\t).'
    )

    @validator('name')
    def check_identifier(cls, v):
        assert all(ord(i) < 128 for i in v), 'Name contains non ASCII characters.'
        assert all(char not in v for char in (',', ';', '!', '\n', '\t')), \
            'Name contains an invalid character for EnergyPlus (, ; ! \\n \\t).'
        return v

    day_type: DesignDayTypes

    dry_bulb_condition: DryBulbCondition = Field(
        ...,
        description='A DryBulbCondition describing temperature conditions on '
            'the design day.'
    )

    humidity_condition: HumidityCondition = Field(
        ...,
        description='A HumidityCondition describing humidity and precipitation '
            'conditions on the design day.'
    )

    wind_condition: WindCondition = Field(
        ...,
        description='A WindCondition describing wind conditions on the design day.'
    )

    sky_condition: Union[ASHRAEClearSky, ASHRAETau]


if __name__ == '__main__':
    print(DesignDay.schema_json(indent=2))



# coding=utf-8
from __future__ import division

from honeybee_radiance.modifierset import ModifierSet
from honeybee_radiance.lib.modifiersets import generic_modifier_set_visible
import honeybee_radiance.lib.modifiers as modifiers

import os
import json


def modifierset_abridged_complete(directory):
    dest_file = os.path.join(directory, 'modifierset_abridged_complete.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_modifier_set_visible.to_dict(True, False), fp, indent=4)


def modifierset_abridged_partial_exterior(directory):
    mod_set = ModifierSet('Generic_Exterior_Visible_Modifier_Set')
    mod_set.wall_set.exterior_modifier = modifiers.generic_wall_exterior
    mod_set.floor_set.exterior_modifier = modifiers.generic_floor_exterior
    mod_set.roof_ceiling_set.exterior_modifier = modifiers.generic_roof_exterior
    dest_file = os.path.join(directory, 'modifierset_abridged_partial_exterior.json')
    with open(dest_file, 'w') as fp:
        json.dump(mod_set.to_dict(True, True), fp, indent=4)


def modifierset_complete(directory):
    dest_file = os.path.join(directory, 'modifierset_complete.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_modifier_set_visible.to_dict(False, False), fp, indent=4)


def modifierset_partial_exterior(directory):
    mod_set = ModifierSet('Generic_Exterior_Visible_Modifier_Set')
    mod_set.wall_set.exterior_modifier = modifiers.generic_wall_exterior
    mod_set.floor_set.exterior_modifier = modifiers.generic_floor_exterior
    mod_set.roof_ceiling_set.exterior_modifier = modifiers.generic_roof_exterior
    dest_file = os.path.join(directory, 'modifierset_partial_exterior.json')
    with open(dest_file, 'w') as fp:
        json.dump(mod_set.to_dict(False, True), fp, indent=4)


# run all functions within the file
master_dir = os.path.split(os.path.dirname(__file__))[0]
sample_directory = os.path.join(master_dir, 'samples', 'modifier_set')

modifierset_abridged_complete(sample_directory)
modifierset_abridged_partial_exterior(sample_directory)
modifierset_complete(sample_directory)
modifierset_partial_exterior(sample_directory)



# TODO: change to Modifers Schema?
"""Material Schema"""
from pydantic import Field, constr, validator, root_validator
from typing import List, Union

from .._base import IDdBaseModel, BaseModel


class Void(BaseModel):
    """Void modifier"""

    type: constr(regex='^void$') = 'void'


class ModifierBase(IDdBaseModel):
    """Base class for Radiance Modifiers"""

    modifier: Union[
         Void, 'Plastic', 'Glass', 'BSDF', 'Glow', 'Light', 'Trans'
    ] = Field(
        default=Void(),
        description='Material modifier (default: Void).'
        )

    dependencies: List[
        Union[Void, 'Plastic', 'Glass', 'BSDF', 'Glow', 'Light', 'Trans']
    ] = Field(
        default=[],
        description='List of modifiers that this modifier depends on. '
                    'This argument is only useful for defining advanced modifiers '
                    'where the modifier is defined based on other modifiers '
                    '(default: []).'
        )


class Plastic(ModifierBase):
    """Radiance plastic material."""

    type: constr(regex='^plastic$') = 'plastic'

    r_reflectance: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the red channel reflectance '
                    '(default: 0).'
    )

    g_reflectance: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the green channel reflectance '
                    '(default: 0).'
    )

    b_reflectance: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the blue channel reflectance '
                    '(default: 0).'
    )

    specularity: float = Field(
        default=0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the fraction of specularity. '
                    'Specularity fractions greater than 0.1 are not realistic. '
                    '(default: 0).'
    )

    roughness: float = Field(
        default=0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the roughness, specified as the '
                    'rms slope of surface facets. Roughness greater than 0.2 are '
                    'not realistic (default: 0).'
    )


class Trans(Plastic):
    """Radiance Translucent material."""

    type: constr(regex='^trans$') = 'trans'

    transmitted_diff: float = Field(
        default=0,
        ge=0,
        le=1,
        description='The fraction of transmitted light that is transmitted diffusely in '
                    'a scattering fashion (default: 0).'
    )

    transmitted_spec: float = Field(
        default=0,
        ge=0,
        le=1,
        description='The fraction of transmitted light that is not diffusely scattered '
                    '(default: 0).'
    )

    @root_validator
    def check_sum_fractions(cls, values):
        """Ensure sum is less than 1."""
        trans_diff = values.get('transmitted_diff')
        trans_spec = values.get('transmitted_spec')
        r_refl = values.get('r_reflectance')
        g_refl = values.get('g_reflectance')
        b_refl = values.get('b_reflectance')
        identifier = values.get('identifier')
        summed = trans_diff + trans_spec + r_refl + g_refl + b_refl
        assert summed <= 1, 'The sum of the transmitted diffuse and specular light ' \
            'fractions cannot be greater than 1, but is {} for modifier {}.'.format(
                summed, identifier)
        return values


class Glass(ModifierBase):
    """Radiance glass material."""

    type: constr(regex='^glass$') = 'glass'

    r_transmissivity: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the red channel transmissivity '
                    '(default: 0).'
    )

    g_transmissivity: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the green channel transmissivity '
                    '(default: 0).'
    )

    b_transmissivity: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the blue channel transmissivity '
                    '(default: 0).'
    )

    refraction_index: float = Field(
        default=1.52,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the index of refraction '
                    '(default: 1.52).'
    )


class BSDF(ModifierBase):
    """Radiance BSDF (Bidirectional Scattering Distribution Function) material."""

    type: constr(regex='^BSDF$') = 'BSDF'

    up_orientation: List[float] = Field(
        default=(0.01, 0.01, 1.00),
        min_items=3,
        max_items=3,
        description='Vector as sequence that sets the hemisphere that the BSDF material '
                    'faces. (default: (0.01, 0.01, 1.00).'
    )

    thickness: float = Field(
        default=0,
        description='Optional number to set the thickness of the BSDF material '
                    'Sign of thickness indicates whether proxied geometry is '
                    'behind the BSDF surface (when thickness is positive) or in '
                    'front (when thickness is negative)(default: 0).'
    )

    function_file: str = Field(
        default='.',
        min_length=1,
        max_length=100,
        description='Optional input for function file (default: ".").'
    )

    transform: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Optional transform input to scale the thickness and reorient '
                    'the up vector (default: None).'
    )

    bsdf_data: bytes = Field(
        ...,
        description='BSDF xml file data as bytes.'
    )

    front_diffuse_reflectance: List[float] = Field(
        default=None,
        min_items=3,
        max_items=3,
        description='Optional additional front diffuse reflectance as sequence of '
                    'numbers (default: None).'
    )

    back_diffuse_reflectance: List[float] = Field(
        default=None,
        min_items=3,
        max_items=3,
        description='Optional additional back diffuse reflectance as sequence of '
                    'numbers (default: None).'
    )

    diffuse_transmittance: List[float] = Field(
        default=None,
        min_items=3,
        max_items=3,
        description='Optional additional diffuse transmittance as sequence of '
                    'numbers (default: None).'
    )

    @validator('front_diffuse_reflectance')
    def check_front_diff_value(cls, values):
        """Ensure every list value is between 0 and 1."""
        assert all(0 <= v <= 1 for v in values), \
            'Every value in front diffuse reflectance must be between 0 and 1.'
        return values

    @validator('back_diffuse_reflectance')
    def check_back_diff_value(cls, values):
        """Ensure every list value is between 0 and 1."""
        assert all(0 <= v <= 1 for v in values), \
            'Every value in back diffuse reflectance must be between 0 and 1.'
        return values

    @validator('diffuse_transmittance')
    def check_diff_trans_value(cls, values):
        """Ensure every list value is between 0 and 1."""
        assert all(0 <= v <= 1 for v in values), \
            'Every value in diffuse transmittance must be between 0 and 1.'
        return values


class Light(ModifierBase):
    """Radiance Light material."""

    type: constr(regex='^light$') = 'light'

    r_emittance: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the red channel of the modifier '
                    '(default: 0).'
    )

    g_emittance: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the green channel of the modifier '
                    '(default: 0).'
    )

    b_emittance: float = Field(
        default=0.0,
        ge=0,
        le=1,
        description='A value between 0 and 1 for the blue channel of the modifier '
                    '(default: 0).'
    )


class Glow(Light):
    """Radiance Glow material."""

    type: constr(regex='^glow$') = 'glow'

    max_radius: float = Field(
        default=0,
        description='Maximum radius for shadow testing (default: 0). Surfaces with zero '
                    'will never be tested for zero, although it may participate in '
                    'interreflection calculation. Negative values will never contribute '
                    'to scene illumination.'
    )


# Required for self.referencing model
# see https://pydantic-docs.helpmanual.io/#self-referencing-models
ModifierBase.update_forward_refs()
Plastic.update_forward_refs()
Glass.update_forward_refs()
BSDF.update_forward_refs()
Glow.update_forward_refs()
Light.update_forward_refs()
Trans.update_forward_refs()




# coding=utf-8
from __future__ import division

from honeybee_energy.construction.window import WindowConstruction
from honeybee_energy.material.shade import EnergyWindowMaterialBlind

from  honeybee_energy.lib.materials import lowe_glass, clear_glass, argon_gap
from honeybee_energy.lib.constructions import generic_exterior_wall, \
    generic_roof, generic_exterior_door, generic_double_pane

import os
import json


def construction_opaque_door(directory):
    dest_file = os.path.join(directory, 'construction_opaque_door.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_exterior_door.to_dict(abridged=True), fp, indent=4)


def construction_opaque_roof(directory):
    dest_file = os.path.join(directory, 'construction_opaque_roof.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_roof.to_dict(abridged=True), fp, indent=4)


def construction_opaque_wall(directory):
    dest_file = os.path.join(directory, 'construction_opaque_wall.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_exterior_wall.to_dict(abridged=True), fp, indent=4)


def construction_window_double(directory):
    dest_file = os.path.join(directory, 'construction_window_double.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_double_pane.to_dict(abridged=True), fp, indent=4)


def construction_window_triple(directory):
    triple_pane = WindowConstruction(
        'Triple Pane Argon', [lowe_glass, argon_gap, lowe_glass, argon_gap, clear_glass])
    dest_file = os.path.join(directory, 'construction_window_triple.json')
    with open(dest_file, 'w') as fp:
        json.dump(triple_pane.to_dict(abridged=True), fp, indent=4)


def construction_window_blinds(directory):
    blinds = EnergyWindowMaterialBlind('Generic Aluminium Blinds')
    window_with_blinds = WindowConstruction(
        'Triple Pane Argon', [lowe_glass, argon_gap, lowe_glass, blinds])
    dest_file = os.path.join(directory, 'construction_window_blinds.json')
    with open(dest_file, 'w') as fp:
        json.dump(window_with_blinds.to_dict(abridged=True), fp, indent=4)


# run all functions within the file
master_dir = os.path.split(os.path.dirname(__file__))[0]
sample_directory = os.path.join(master_dir, 'samples', 'construction')

construction_opaque_door(sample_directory)
construction_opaque_roof(sample_directory)
construction_opaque_wall(sample_directory)
construction_window_double(sample_directory)
construction_window_triple(sample_directory)
construction_window_blinds(sample_directory)



from honeybee_schema.energy.programtype import ProgramTypeAbridged, ProgramType
import os

# target folder where all of the samples live
root = os.path.dirname(os.path.dirname(__file__))
target_folder = os.path.join(root, 'samples', 'program_type')


def test_program_type_abridged_plenum():
    file_path = os.path.join(target_folder, 'program_type_abridged_plenum.json')
    ProgramTypeAbridged.parse_file(file_path)


def test_program_type_abridged_office():
    file_path = os.path.join(target_folder, 'program_type_abridged_office.json')
    ProgramTypeAbridged.parse_file(file_path)


def test_program_type_abridged_kitchen():
    file_path = os.path.join(target_folder, 'program_type_abridged_kitchen.json')
    ProgramTypeAbridged.parse_file(file_path)


def test_program_type_abridged_patient_room():
    file_path = os.path.join(target_folder, 'program_type_abridged_patient_room.json')
    ProgramTypeAbridged.parse_file(file_path)


def test_program_type_plenum():
    file_path = os.path.join(target_folder, 'program_type_plenum.json')
    ProgramType.parse_file(file_path)


def test_program_type_office():
    file_path = os.path.join(target_folder, 'program_type_office.json')
    ProgramType.parse_file(file_path)



# coding=utf-8
from __future__ import division

from honeybee_energy.schedule.typelimit import ScheduleTypeLimit
from honeybee_energy.schedule.day import ScheduleDay
from honeybee_energy.schedule.rule import ScheduleRule
from honeybee_energy.schedule.ruleset import ScheduleRuleset
from honeybee_energy.schedule.fixedinterval import ScheduleFixedInterval
from honeybee_energy.lib.schedules import schedule_by_identifier
import honeybee_energy.lib.scheduletypelimits as schedule_types

from ladybug.dt import Date, Time

import random
import os
import json


def scheduletypelimit_temperature(directory):
    temperature = ScheduleTypeLimit('Temperature', -273.15, None, 'Continuous', 'Temperature')

    dest_file = os.path.join(directory, 'scheduletypelimit_temperature.json')
    with open(dest_file, 'w') as fp:
        json.dump(temperature.to_dict(), fp, indent=4)


def schedule_ruleset_simple_repeating(directory):
    test_vals = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 1, 1, 1, 1, 1, 1,
                 0.5, 0.5, 0.5, 0.5, 1, 1, 1, 1, 0.5, 0.5, 0.5, 0.5]
    sched = ScheduleRuleset.from_daily_values('Simple Repeating', test_vals)

    dest_file = os.path.join(directory, 'schedule_ruleset_simple_repeating.json')
    with open(dest_file, 'w') as fp:
        json.dump(sched.to_dict(True), fp, indent=4)


def schedule_primary_school_occupancy(directory):
    weekday_school = ScheduleDay('Weekday School Year', [0, 1, 0.5, 0],
                                 [Time(0, 0), Time(8, 0), Time(15, 0), Time(18, 0)])
    weekend_school = ScheduleDay('Weekend School Year', [0])
    weekday_summer = ScheduleDay('Weekday Summer', [0, 0.5, 0],
                                 [Time(0, 0), Time(9, 0), Time(17, 0)])
    weekend_summer = ScheduleDay('Weekend Summer', [0])

    summer_weekday_rule = ScheduleRule(
        weekday_summer, start_date=Date(7, 1), end_date=Date(9, 1))
    summer_weekday_rule.apply_weekday = True
    summer_weekend_rule = ScheduleRule(
        weekend_summer, start_date=Date(7, 1), end_date=Date(9, 1))
    summer_weekend_rule.apply_weekend = True
    school_weekend_rule = ScheduleRule(weekend_school)
    school_weekend_rule.apply_weekend = True

    summer_design = ScheduleDay('School Summer Design', [0, 1, 0.25],
                                [Time(0, 0), Time(6, 0), Time(18, 0)])
    winter_design = ScheduleDay('School Winter Design', [0])

    schedule = ScheduleRuleset('School Occupancy', weekday_school,
                               [summer_weekday_rule, summer_weekend_rule,
                                school_weekend_rule], schedule_types.fractional,
                                weekend_summer, summer_design, winter_design)

    dest_file = os.path.join(directory, 'schedule_primary_school_occupancy.json')
    with open(dest_file, 'w') as fp:
        json.dump(schedule.to_dict(True), fp, indent=4)


def schedule_ruleset_office_occupancy(directory):
    office_sch = schedule_by_identifier('OfficeLarge BLDG_OCC_SCH')
    dest_file = os.path.join(directory, 'schedule_ruleset_office_occupancy.json')
    with open(dest_file, 'w') as fp:
        json.dump(office_sch.to_dict(True), fp, indent=4)


def schedule_fixedinterval_increasing_single_day(directory):
    increase_sched = ScheduleFixedInterval(
        'Solstice Increasing', [round(x / 23, 4) for x in range(24)],
        schedule_types.fractional, start_date=Date(6, 21))

    dest_file = os.path.join(
        directory, 'schedule_fixedinterval_increasing_single_day.json')
    with open(dest_file, 'w') as fp:
        json.dump(increase_sched.to_dict(True), fp, indent=4)


def schedule_fixedinterval_increasing_fine_timestep(directory):
    increase_sched = ScheduleFixedInterval(
        'Solstice Increasing', [round(x / 143, 4) for x in range(144)],
        schedule_types.fractional, start_date=Date(6, 21), timestep=6,)

    dest_file = os.path.join(
        directory, 'schedule_fixedinterval_increasing_fine_timestep.json')
    with open(dest_file, 'w') as fp:
        json.dump(increase_sched.to_dict(True), fp, indent=4)


def schedule_fixedinterval_random_annual(directory):
    occ_sched = ScheduleFixedInterval(
        'Random Occupancy', [round(random.random(), 4) for i in range(8760)],
        schedule_types.fractional)

    dest_file = os.path.join(
        directory, 'schedule_fixedinterval_random_annual.json')
    with open(dest_file, 'w') as fp:
        json.dump(occ_sched.to_dict(True), fp, indent=4)


def schedule_fixedinterval_leap_year(directory):
    test_vals = [15, 15, 15, 15, 15, 15, 20, 20, 20, 20, 20, 20,
                 20, 20, 20, 20, 20, 20, 20, 20, 15, 15, 15, 15]
    sched = ScheduleFixedInterval(
        'Weekly Temperature', test_vals * 7,
        schedule_types.temperature, start_date=Date(2, 29, True))

    dest_file = os.path.join(directory, 'schedule_fixedinterval_leap_year.json')
    with open(dest_file, 'w') as fp:
        json.dump(sched.to_dict(True), fp, indent=4)


# run all functions within the file
master_dir = os.path.split(os.path.dirname(__file__))[0]
sample_directory = os.path.join(master_dir, 'samples', 'schedule')

scheduletypelimit_temperature(sample_directory)
schedule_ruleset_office_occupancy(sample_directory)
schedule_ruleset_simple_repeating(sample_directory)
schedule_primary_school_occupancy(sample_directory)
schedule_fixedinterval_increasing_single_day(sample_directory)
schedule_fixedinterval_increasing_fine_timestep(sample_directory)
schedule_fixedinterval_random_annual(sample_directory)
schedule_fixedinterval_leap_year(sample_directory)



# coding=utf-8
from honeybee.model import Model
from honeybee.room import Room
from honeybee.face import Face
from honeybee.shade import Shade
from honeybee.aperture import Aperture
from honeybee.door import Door
from honeybee.boundarycondition import boundary_conditions
from honeybee.facetype import face_types, Floor, RoofCeiling

from honeybee_energy.constructionset import ConstructionSet
from honeybee_energy.construction.opaque import OpaqueConstruction
from honeybee_energy.construction.window import WindowConstruction
from honeybee_energy.construction.shade import ShadeConstruction
from honeybee_energy.construction.air import AirBoundaryConstruction
from honeybee_energy.material.opaque import EnergyMaterial
from honeybee_energy.schedule.fixedinterval import ScheduleFixedInterval

import honeybee_energy.lib.programtypes as prog_type_lib
import honeybee_energy.lib.scheduletypelimits as schedule_types
from honeybee_energy.lib.materials import clear_glass, air_gap, roof_membrane, \
    wood, insulation
from honeybee_energy.lib.constructions import generic_exterior_wall, \
    generic_interior_wall, generic_interior_floor, generic_interior_ceiling, \
    generic_double_pane

from honeybee_radiance.modifierset import ModifierSet
from honeybee_radiance.modifier.material import Glass, Plastic

from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.plane import Plane
from ladybug_geometry.geometry3d.face import Face3D
from ladybug_geometry.geometry3d.polyface import Polyface3D

import os
import json
import random


def model_complete_single_zone_office(directory):
    room = Room.from_box('Tiny_House_Office', 5, 10, 3)
    room.properties.energy.program_type = prog_type_lib.office_program
    room.properties.energy.add_default_ideal_air()

    stone = EnergyMaterial('Thick Stone', 0.3, 2.31, 2322, 832, 'Rough',
                           0.95, 0.75, 0.8)
    thermal_mass_constr = OpaqueConstruction('Thermal Mass Floor', [stone])
    room[0].properties.energy.construction = thermal_mass_constr

    south_face = room[3]
    south_face.apertures_by_ratio(0.4, 0.01)
    south_face.apertures[0].overhang(0.5, indoor=False)
    south_face.apertures[0].overhang(0.5, indoor=True)
    south_face.move_shades(Vector3D(0, 0, -0.5))
    light_shelf_out = ShadeConstruction('Outdoor_Light_Shelf', 0.5, 0.5)
    light_shelf_in = ShadeConstruction('Indoor_Light_Shelf', 0.7, 0.7)
    south_face.apertures[0].outdoor_shades[0].properties.energy.construction = light_shelf_out
    south_face.apertures[0].indoor_shades[0].properties.energy.construction = light_shelf_in

    north_face = room[1]
    north_face.overhang(0.25, indoor=False)
    door_verts = [Point3D(2, 10, 0.1), Point3D(1, 10, 0.1),
                  Point3D(1, 10, 2.5), Point3D(2, 10, 2.5)]
    door = Door('Front_Door', Face3D(door_verts))
    north_face.add_door(door)

    aperture_verts = [Point3D(4.5, 10, 1), Point3D(2.5, 10, 1),
                      Point3D(2.5, 10, 2.5), Point3D(4.5, 10, 2.5)]
    aperture = Aperture('Front_Aperture', Face3D(aperture_verts))
    triple_pane = WindowConstruction(
        'Triple Pane Window', [clear_glass, air_gap, clear_glass, air_gap, clear_glass])
    aperture.properties.energy.construction = triple_pane
    north_face.add_aperture(aperture)

    tree_canopy_geo = Face3D.from_regular_polygon(
        6, 2, Plane(Vector3D(0, 0, 1), Point3D(5, -3, 4)))
    tree_canopy = Shade('Tree_Canopy', tree_canopy_geo)

    table_geo = Face3D.from_rectangle(2, 2, Plane(o=Point3D(1.5, 4, 1)))
    table = Shade('Table', table_geo)
    room.add_indoor_shade(table)

    model = Model('Tiny_House', [room], orphaned_shades=[tree_canopy])
    model.north_angle = 15

    model_dict = model.to_dict()

    dest_file = os.path.join(directory, 'model_complete_single_zone_office.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(), fp, indent=4)


def model_complete_single_zone_office_user_data(directory):
    room = Room.from_box('Tiny_House_Office', 5, 10, 3)
    room.properties.energy.program_type = prog_type_lib.office_program
    room.properties.energy.add_default_ideal_air()

    south_face = room[3]
    south_face.apertures_by_ratio(0.4, 0.01)
    south_face.apertures[0].overhang(0.5, indoor=False)
    south_face.apertures[0].overhang(0.5, indoor=True)
    south_face.move_shades(Vector3D(0, 0, -0.5))
    light_shelf_out = ShadeConstruction('Outdoor_Light_Shelf', 0.5, 0.5)
    light_shelf_in = ShadeConstruction('Indoor_Light_Shelf', 0.7, 0.7)
    south_face.apertures[0].outdoor_shades[0].properties.energy.construction = light_shelf_out
    south_face.apertures[0].indoor_shades[0].properties.energy.construction = light_shelf_in

    north_face = room[1]
    north_face.overhang(0.25, indoor=False)
    door_verts = [Point3D(2, 10, 0.1), Point3D(1, 10, 0.1),
                  Point3D(1, 10, 2.5), Point3D(2, 10, 2.5)]
    door = Door('Front_Door', Face3D(door_verts))
    north_face.add_door(door)

    aperture_verts = [Point3D(4.5, 10, 1), Point3D(2.5, 10, 1),
                      Point3D(2.5, 10, 2.5), Point3D(4.5, 10, 2.5)]
    aperture = Aperture('Front_Aperture', Face3D(aperture_verts))
    north_face.add_aperture(aperture)
    
    model = Model('Tiny_House', [room])
    model_dict = model.to_dict()

    model_dict['user_data'] = {'site': 'The backyard'}
    model_dict['rooms'][0]['user_data'] = {'alt_name': 'Little old tiny house'}
    model_dict['rooms'][0]['faces'][0]['user_data'] = {'alt_name': 'The floor'}
    model_dict['rooms'][0]['faces'][3]['apertures'][0]['user_data'] = \
        {'alt_name': 'Picture window'}
    model_dict['rooms'][0]['faces'][1]['doors'][0]['user_data'] = \
        {'alt_name': 'Front door'}
    model_dict['rooms'][0]['faces'][3]['apertures'][0]['outdoor_shades'][0]['user_data'] = \
        {'alt_name': 'Awning'}

    dest_file = os.path.join(directory, 'model_complete_user_data.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(), fp, indent=4)


def model_complete_multi_zone_office(directory):
    first_floor = Room.from_box('First_Floor', 10, 10, 3, origin=Point3D(0, 0, 0))
    second_floor = Room.from_box('Second_Floor', 10, 10, 3, origin=Point3D(0, 0, 3))
    first_floor.properties.energy.program_type = prog_type_lib.office_program
    second_floor.properties.energy.program_type = prog_type_lib.office_program
    first_floor.properties.energy.add_default_ideal_air()
    second_floor.properties.energy.add_default_ideal_air()
    for face in first_floor[1:5]:
        face.apertures_by_ratio(0.2, 0.01)
    for face in second_floor[1:5]:
        face.apertures_by_ratio(0.2, 0.01)

    pts_1 = [Point3D(0, 0, 6), Point3D(0, 10, 6), Point3D(10, 10, 6), Point3D(10, 0, 6)]
    pts_2 = [Point3D(0, 0, 6), Point3D(5, 0, 9), Point3D(5, 10, 9), Point3D(0, 10, 6)]
    pts_3 = [Point3D(10, 0, 6), Point3D(10, 10, 6), Point3D(5, 10, 9), Point3D(5, 0, 9)]
    pts_4 = [Point3D(0, 0, 6), Point3D(10, 0, 6), Point3D(5, 0, 9)]
    pts_5 = [Point3D(10, 10, 6), Point3D(0, 10, 6), Point3D(5, 10, 9)]
    face_1 = Face('AtticFace1', Face3D(pts_1))
    face_2 = Face('AtticFace2', Face3D(pts_2))
    face_3 = Face('AtticFace3', Face3D(pts_3))
    face_4 = Face('AtticFace4', Face3D(pts_4))
    face_5 = Face('AtticFace5', Face3D(pts_5))
    attic = Room('Attic', [face_1, face_2, face_3, face_4, face_5], 0.01, 1)

    constr_set = ConstructionSet('Attic Construction Set')
    polyiso = EnergyMaterial('PolyIso', 0.2, 0.03, 43, 1210, 'MediumRough')
    roof_constr = OpaqueConstruction('Attic Roof Construction',
                                     [roof_membrane, polyiso, wood])
    floor_constr = OpaqueConstruction('Attic Floor Construction',
                                      [wood, insulation, wood])
    constr_set.floor_set.interior_construction = floor_constr
    constr_set.roof_ceiling_set.exterior_construction = roof_constr
    attic.properties.energy.construction_set = constr_set

    Room.solve_adjacency([first_floor, second_floor, attic], 0.01)

    model = Model('Multi_Zone_Single_Family_House', [first_floor, second_floor, attic])

    dest_file = os.path.join(directory, 'model_complete_multi_zone_office.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(), fp, indent=4)


def model_complete_multiroom_radiance(directory):
    triple_pane = Glass.from_single_transmittance('Triple_Pane_0.35', 0.35)
    first_floor = Room.from_box('First_Floor', 10, 10, 3, origin=Point3D(0, 0, 0))
    second_floor = Room.from_box('Second_Floor', 10, 10, 3, origin=Point3D(0, 0, 3))
    for face in first_floor[1:5]:
        face.apertures_by_ratio(0.2, 0.01)
        face.apertures[0].properties.radiance.modifier = triple_pane
    for face in second_floor[1:5]:
        face.apertures_by_ratio(0.2, 0.01)

    pts_1 = [Point3D(0, 0, 6), Point3D(0, 10, 6), Point3D(10, 10, 6), Point3D(10, 0, 6)]
    pts_2 = [Point3D(0, 0, 6), Point3D(5, 0, 9), Point3D(5, 10, 9), Point3D(0, 10, 6)]
    pts_3 = [Point3D(10, 0, 6), Point3D(10, 10, 6), Point3D(5, 10, 9), Point3D(5, 0, 9)]
    pts_4 = [Point3D(0, 0, 6), Point3D(10, 0, 6), Point3D(5, 0, 9)]
    pts_5 = [Point3D(10, 10, 6), Point3D(0, 10, 6), Point3D(5, 10, 9)]
    face_1 = Face('AtticFace1', Face3D(pts_1))
    face_2 = Face('AtticFace2', Face3D(pts_2))
    face_3 = Face('AtticFace3', Face3D(pts_3))
    face_4 = Face('AtticFace4', Face3D(pts_4))
    face_5 = Face('AtticFace5', Face3D(pts_5))
    attic = Room('Attic', [face_1, face_2, face_3, face_4, face_5], 0.01, 1)

    mod_set = ModifierSet('Attic_Modifier_Set')
    polyiso = Plastic.from_single_reflectance('PolyIso', 0.45)
    mod_set.roof_ceiling_set.exterior_modifier = polyiso
    attic.properties.radiance.modifier_set = mod_set

    Room.solve_adjacency([first_floor, second_floor, attic], 0.01)

    model = Model('Multi_Room_Radiance_House', [first_floor, second_floor, attic])

    dest_file = os.path.join(directory, 'model_complete_multiroom_radiance.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(), fp, indent=4)


def model_complete_patient_room(directory):
    pat_room_program = \
        prog_type_lib.program_type_by_identifier('2013::Hospital::ICU_PatRm')
    room = Room.from_box('Hospital_Patient_Room', 5, 10, 3)
    room.properties.energy.program_type = pat_room_program

    room.properties.energy.add_default_ideal_air()
    ideal_air = room.properties.energy.hvac.duplicate()
    ideal_air.economizer_type = 'DifferentialEnthalpy'
    ideal_air.sensible_heat_recovery = 0.81
    ideal_air.latent_heat_recovery = 0.68
    room.properties.energy.hvac = ideal_air

    pat_rm_setpoint = room.properties.energy.setpoint.duplicate()
    pat_rm_setpoint.identifier = 'Humidity Controlled PatRm Setpt'
    pat_rm_setpoint.heating_setpoint = 21
    pat_rm_setpoint.cooling_setpoint = 24
    pat_rm_setpoint.humidifying_setpoint = 30
    pat_rm_setpoint.dehumidifying_setpoint = 55
    room.properties.energy.setpoint = pat_rm_setpoint

    south_face = room[3]
    south_face.apertures_by_ratio(0.4, 0.01)
    south_face.apertures[0].overhang(0.5, indoor=False)
    south_face.move_shades(Vector3D(0, 0, -0.5))

    room[0].boundary_condition = boundary_conditions.adiabatic
    room[1].boundary_condition = boundary_conditions.adiabatic
    room[2].boundary_condition = boundary_conditions.adiabatic
    room[4].boundary_condition = boundary_conditions.adiabatic
    room[5].boundary_condition = boundary_conditions.adiabatic

    model = Model('Patient_Room_Test_Box', [room])

    dest_file = os.path.join(directory, 'model_complete_patient_room.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(), fp, indent=4)


def model_complete_office_floor(directory):
    pts_1 = [Point3D(0, 0), Point3D(30, 0), Point3D(20, 10), Point3D(10, 10)]
    pts_2 = [Point3D(0, 0), Point3D(10, 10), Point3D(10, 20), Point3D(0, 30)]
    pts_3 = [Point3D(10, 20), Point3D(20, 20), Point3D(30, 30), Point3D(0, 30)]
    pts_4 = [Point3D(30, 0), Point3D(30, 30), Point3D(20, 20), Point3D(20, 10)]
    verts = [pts_1, pts_2, pts_3, pts_4]
    rooms = []
    for i, f_vert in enumerate(verts):
        pface = Polyface3D.from_offset_face(Face3D(f_vert), 3)
        room = Room.from_polyface3d('PerimeterRoom{}'.format(i), pface)
        room.properties.energy.program_type = prog_type_lib.office_program
        room.properties.energy.add_default_ideal_air()
        rooms.append(room)
    rooms.append(Room.from_box('CoreRoom', 10, 10, 3, origin=Point3D(10, 10)))
    adj_info = Room.solve_adjacency(rooms, 0.01)
    for face_pair in adj_info['adjacent_faces']:
        face_pair[0].type = face_types.air_boundary
        face_pair[1].type = face_types.air_boundary
    for room in rooms:
        for face in room:
            if isinstance(face.type, (Floor, RoofCeiling)):
                face.boundary_condition = boundary_conditions.adiabatic

    model = Model('Core_Perimeter_Office_Floor', rooms)

    dest_file = os.path.join(directory, 'model_complete_office_floor.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(), fp, indent=4)


def model_energy_shoe_box(directory):
    room = Room.from_box('Simple_Shoe_Box_Zone', 5, 10, 3)
    room[0].boundary_condition = boundary_conditions.adiabatic
    for face in room[2:]:
        face.boundary_condition = boundary_conditions.adiabatic

    north_face = room[1]
    north_face.apertures_by_ratio_rectangle(0.4, 2, 0.7, 2, 0, 0.01)

    constr_set = ConstructionSet('Shoe Box Construction Set')
    constr_set.wall_set.exterior_construction = generic_exterior_wall
    constr_set.wall_set.interior_construction = generic_interior_wall
    constr_set.floor_set.interior_construction = generic_interior_floor
    constr_set.roof_ceiling_set.interior_construction = generic_interior_ceiling
    constr_set.aperture_set.window_construction = generic_double_pane
    room.properties.energy.construction_set = constr_set

    model = Model('Shoe_Box', [room])

    dest_file = os.path.join(directory, 'model_energy_shoe_box.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(included_prop=['energy']), fp, indent=4)


def model_energy_detailed_loads(directory):
    room = Room.from_box('Office_Test_Box', 5, 10, 3)
    room.properties.energy.program_type = prog_type_lib.plenum_program
    room.properties.energy.add_default_ideal_air()

    room.properties.energy.people = prog_type_lib.office_program.people
    room.properties.energy.lighting = prog_type_lib.office_program.lighting
    room.properties.energy.electric_equipment = prog_type_lib.office_program.electric_equipment
    room.properties.energy.infiltration = prog_type_lib.office_program.infiltration
    room.properties.energy.ventilation = prog_type_lib.office_program.ventilation
    room.properties.energy.setpoint = prog_type_lib.office_program.setpoint

    room[0].boundary_condition = boundary_conditions.adiabatic
    room[1].boundary_condition = boundary_conditions.adiabatic
    room[2].boundary_condition = boundary_conditions.adiabatic
    room[4].boundary_condition = boundary_conditions.adiabatic
    room[5].boundary_condition = boundary_conditions.adiabatic

    model = Model('Office_Model', [room])

    dest_file = os.path.join(
        directory, 'model_energy_detailed_loads.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(included_prop=['energy']), fp, indent=4)


def model_energy_fixed_interval(directory):
    room = Room.from_box('Tiny_House_Office', 5, 10, 3)
    room.properties.energy.program_type = prog_type_lib.office_program
    room.properties.energy.add_default_ideal_air()

    occ_sched = ScheduleFixedInterval(
        'Random Occupancy', [round(random.random(), 4) for i in range(8760)],
        schedule_types.fractional)
    new_people = room.properties.energy.people.duplicate()
    new_people.occupancy_schedule = occ_sched
    room.properties.energy.people = new_people

    south_face = room[3]
    south_face.apertures_by_ratio(0.4, 0.01)
    south_face.apertures[0].overhang(0.5, indoor=False)
    south_face.apertures[0].overhang(0.5, indoor=True)
    south_face.move_shades(Vector3D(0, 0, -0.5))
    light_shelf_out = ShadeConstruction('Outdoor_Light_Shelf', 0.5, 0.5)
    light_shelf_in = ShadeConstruction('Indoor_Light_Shelf', 0.7, 0.7)
    south_face.apertures[0].outdoor_shades[0].properties.energy.construction = light_shelf_out
    south_face.apertures[0].indoor_shades[0].properties.energy.construction = light_shelf_in

    north_face = room[1]
    north_face.overhang(0.25, indoor=False)
    door_verts = [Point3D(2, 10, 0.1), Point3D(1, 10, 0.1),
                  Point3D(1, 10, 2.5), Point3D(2, 10, 2.5)]
    door = Door('Front_Door', Face3D(door_verts))
    north_face.add_door(door)

    aperture_verts = [Point3D(4.5, 10, 1), Point3D(2.5, 10, 1),
                      Point3D(2.5, 10, 2.5), Point3D(4.5, 10, 2.5)]
    aperture = Aperture('Front_Aperture', Face3D(aperture_verts))
    north_face.add_aperture(aperture)

    tree_canopy_geo = Face3D.from_regular_polygon(
        6, 2, Plane(Vector3D(0, 0, 1), Point3D(5, -3, 4)))
    tree_canopy = Shade('Tree_Canopy', tree_canopy_geo)
    winter = [0.75] * 2190
    spring = [0.75 - ((x / 2190) * 0.5) for x in range(2190)]
    summer = [0.25] * 2190
    fall = [0.25 + ((x / 2190) * 0.5) for x in range(2190)]
    trans_sched = ScheduleFixedInterval(
        'Seasonal Tree Transmittance', winter + spring + summer + fall,
        schedule_types.fractional)
    tree_canopy.properties.energy.transmittance_schedule = trans_sched

    model = Model('Tiny_House', [room], orphaned_shades=[tree_canopy])
    model.north_angle = 15

    dest_file = os.path.join(
        directory, 'model_energy_fixed_interval.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(included_prop=['energy']), fp, indent=4)
    

def model_energy_no_program(directory):
    room = Room.from_box('Abandoned_Tiny_House', 5, 10, 3)

    stone = EnergyMaterial('Thick Stone', 0.3, 2.31, 2322, 832, 'Rough',
                           0.95, 0.75, 0.8)
    thermal_mass_constr = OpaqueConstruction('Thermal Mass Floor', [stone])
    room[0].properties.energy.construction = thermal_mass_constr

    south_face = room[3]
    south_face.apertures_by_ratio(0.4, 0.01)
    south_face.apertures[0].overhang(0.5, indoor=False)
    south_face.apertures[0].overhang(0.5, indoor=True)
    south_face.move_shades(Vector3D(0, 0, -0.5))
    light_shelf_out = ShadeConstruction('Outdoor_Light_Shelf', 0.5, 0.5)
    light_shelf_in = ShadeConstruction('Indoor_Light_Shelf', 0.7, 0.7)
    south_face.apertures[0].outdoor_shades[0].properties.energy.construction = light_shelf_out
    south_face.apertures[0].indoor_shades[0].properties.energy.construction = light_shelf_in

    north_face = room[1]
    north_face.overhang(0.25, indoor=False)
    door_verts = [Point3D(2, 10, 0.1), Point3D(1, 10, 0.1),
                  Point3D(1, 10, 2.5), Point3D(2, 10, 2.5)]
    door = Door('Front_Door', Face3D(door_verts))
    north_face.add_door(door)

    aperture_verts = [Point3D(4.5, 10, 1), Point3D(2.5, 10, 1),
                      Point3D(2.5, 10, 2.5), Point3D(4.5, 10, 2.5)]
    aperture = Aperture('Front_Aperture', Face3D(aperture_verts))
    triple_pane = WindowConstruction(
        'Triple Pane Window', [clear_glass, air_gap, clear_glass, air_gap, clear_glass])
    aperture.properties.energy.construction = triple_pane
    north_face.add_aperture(aperture)

    tree_canopy_geo = Face3D.from_regular_polygon(
        6, 2, Plane(Vector3D(0, 0, 1), Point3D(5, -3, 4)))
    tree_canopy = Shade('Tree_Canopy', tree_canopy_geo)

    table_geo = Face3D.from_rectangle(2, 2, Plane(o=Point3D(1.5, 4, 1)))
    table = Shade('Table', table_geo)
    room.add_indoor_shade(table)

    model = Model('Tiny_House', [room], orphaned_shades=[tree_canopy])
    model.north_angle = 15

    dest_file = os.path.join(directory, 'model_energy_no_program.json')
    with open(dest_file, 'w') as fp:
        json.dump(model.to_dict(included_prop=['energy']), fp, indent=4)


def model_energy_properties_office(directory):
    room = Room.from_box('Closed_Office', 5, 10, 3)
    room.properties.energy.program_type = prog_type_lib.office_program
    room.properties.energy.add_default_ideal_air()

    model = Model('Office_Test_Box', [room])
    model_dict = model.to_dict()

    dest_file = os.path.join(directory, 'model_energy_properties_office.json')
    with open(dest_file, 'w') as fp:
        json.dump(model_dict['properties']['energy'], fp, indent=4)


# run all functions within the file
master_dir = os.path.split(os.path.dirname(__file__))[0]
sample_directory = os.path.join(master_dir, 'samples', 'model')

model_complete_single_zone_office(sample_directory)
model_complete_single_zone_office_user_data(sample_directory)
model_complete_multi_zone_office(sample_directory)
model_complete_patient_room(sample_directory)
model_complete_office_floor(sample_directory)

model_complete_multiroom_radiance(sample_directory)

model_energy_shoe_box(sample_directory)
model_energy_detailed_loads(sample_directory)
model_energy_fixed_interval(sample_directory)
model_energy_no_program(sample_directory)
model_energy_properties_office(sample_directory)



from honeybee_schema.energy.construction import OpaqueConstructionAbridged, \
        WindowConstructionAbridged
from copy import copy
from pydantic import ValidationError
import pytest
import os
import json

# target folder where all of the samples live
root = os.path.dirname(os.path.dirname(__file__))
target_folder = os.path.join(root, 'samples', 'construction')


def test_construction_window_double():
    file_path = os.path.join(target_folder, 'construction_window_double.json')
    WindowConstructionAbridged.parse_file(file_path)


def test_construction_window_triple():
    file_path = os.path.join(target_folder, 'construction_window_triple.json')
    WindowConstructionAbridged.parse_file(file_path)


def test_construction_window_blinds():
    file_path = os.path.join(target_folder, 'construction_window_blinds.json')
    WindowConstructionAbridged.parse_file(file_path)


def test_construction_opaque_door():
    file_path = os.path.join(target_folder, 'construction_opaque_door.json')
    OpaqueConstructionAbridged.parse_file(file_path)


def test_construction_opaque_roof():
    file_path = os.path.join(target_folder, 'construction_opaque_roof.json')
    OpaqueConstructionAbridged.parse_file(file_path)


def test_construction_opaque_wall():
    file_path = os.path.join(target_folder, 'construction_opaque_wall.json')
    OpaqueConstructionAbridged.parse_file(file_path)


def test_length_opaque():
    file_path = os.path.join(target_folder, 'construction_opaque_wall.json')
    with open(file_path) as json_file:
        construction_wall = json.load(json_file)
    cons_length_test = copy(construction_wall)
    for i in range(10):
        cons_length_test['layers'].append('material_{}'.format(i))
    with pytest.raises(ValidationError):
        OpaqueConstructionAbridged.parse_obj(cons_length_test)
    cons_length_test['layers'] = []
    with pytest.raises(ValidationError):
        OpaqueConstructionAbridged.parse_obj(cons_length_test)


def test_length_window():
    file_path = os.path.join(target_folder, 'construction_window_double.json')
    with open(file_path) as json_file:
        construction_window = json.load(json_file)
    cons_length_test = copy(construction_window)
    for i in range(8):
        cons_length_test['layers'].append('material_{}'.format(i))
    with pytest.raises(ValidationError):
        WindowConstructionAbridged.parse_obj(cons_length_test)
    cons_length_test['layers'] = []
    with pytest.raises(ValidationError):
        WindowConstructionAbridged.parse_obj(cons_length_test)



"""Objects used as alternatives to numerical properties."""
from pydantic import constr
from ._base import NoExtraBaseModel


class NoLimit(NoExtraBaseModel):

    type: constr(regex='^NoLimit$') = 'NoLimit'


class Autocalculate(NoExtraBaseModel):

    type: constr(regex='^Autocalculate$') = 'Autocalculate'


class Autosize(NoExtraBaseModel):

    type: constr(regex='^Autosize$') = 'Autosize'



"""Programtype Schema"""
from pydantic import Field, constr
from typing import List, Union

from ._base import IDdEnergyBaseModel
from .schedule import ScheduleRuleset, ScheduleFixedInterval
from .load import PeopleAbridged, LightingAbridged, ElectricEquipmentAbridged, \
    GasEquipmentAbridged, InfiltrationAbridged, VentilationAbridged, SetpointAbridged, \
    People, Lighting, ElectricEquipment, GasEquipment, Infiltration, Ventilation, \
    Setpoint



class ProgramTypeAbridged(IDdEnergyBaseModel):

    type: constr(regex='^ProgramTypeAbridged$') = 'ProgramTypeAbridged'

    people: PeopleAbridged = Field(
        default = None,
        description='People to describe the occupancy of the program. If None, '
            'no occupancy will be assumed for the program.'
    )

    lighting: LightingAbridged = Field(
        default = None,
        description='Lighting to describe the lighting usage of the program. '
            'If None, no lighting will be assumed for the program.'
    )

    electric_equipment: ElectricEquipmentAbridged = Field(
        default = None,
        description='ElectricEquipment to describe the usage of electric equipment '
            'within the program. If None, no electric equipment will be assumed.'
    )

    gas_equipment: GasEquipmentAbridged = Field(
        default = None,
        description='GasEquipment to describe the usage of gas equipment '
            'within the program. If None, no gas equipment will be assumed.'
    )

    infiltration : InfiltrationAbridged = Field(
        default = None,
        description='Infiltration to describe the outdoor air leakage of '
            'the program. If None, no infiltration will be assumed for the program.'
    )

    ventilation : VentilationAbridged = Field(
        default = None,
        description='Ventilation to describe the minimum outdoor air requirement '
            'of the program. If None, no ventilation requirement will be assumed.'
    )

    setpoint: SetpointAbridged = Field(
        default = None,
        description='Setpoint object to describe the temperature and humidity setpoints '
            'of the program.  If None, the ProgramType cannot be assigned to a Room '
            'that is conditioned.'
    )


class ProgramType(ProgramTypeAbridged):

    type: constr(regex='^ProgramType$') = 'ProgramType'

    people: People = Field(
        default=None,
        description='People to describe the occupancy of the program. If None, '
            'no occupancy will be assumed for the program.'
    )

    lighting: Lighting = Field(
        default=None,
        description='Lighting to describe the lighting usage of the program. '
            'If None, no lighting will be assumed for the program.'
    )

    electric_equipment: ElectricEquipment = Field(
        default=None,
        description='ElectricEquipment to describe the usage of electric equipment '
            'within the program. If None, no electric equipment will be assumed.'
    )

    gas_equipment: GasEquipment = Field(
        default=None,
        description='GasEquipment to describe the usage of gas equipment '
            'within the program. If None, no gas equipment will be assumed.'
    )

    infiltration: Infiltration = Field(
        default=None,
        description='Infiltration to describe the outdoor air leakage of '
            'the program. If None, no infiltration will be assumed for the program.'
    )

    ventilation: Ventilation = Field(
        default=None,
        description='Ventilation to describe the minimum outdoor air requirement '
            'of the program. If None, no ventilation requirement will be assumed.'
    )

    setpoint: Setpoint = Field(
        default=None,
        description='Setpoint object to describe the temperature and humidity setpoints '
            'of the program.  If None, the ProgramType cannot be assigned to a Room '
            'that is conditioned.'
    )


if __name__ == '__main__':
    print(ProgramTypeAbridged.schema_json(indent=2))
    print(ProgramType.schema_json(indent=2))



"""Schedule Type Limit Schema"""
from pydantic import Field, validator, root_validator, constr, conlist
from typing import List, Union
from enum import Enum
import datetime

from ._base import IDdEnergyBaseModel, DatedBaseModel
from ..altnumber import NoLimit


class ScheduleNumericType (str, Enum):
    """Designates how the range values are validated."""
    continuous = 'Continuous'
    discrete = 'Discrete'


class ScheduleUnitType (str, Enum):
    dimensionless = 'Dimensionless'
    temperature = 'Temperature'
    delta_temperature = 'DeltaTemperature'
    precipitation_rate = 'PrecipitationRate'
    angle = 'Angle'
    convection_coefficient = 'ConvectionCoefficient'
    activity_level = 'ActivityLevel'
    velocity = 'Velocity'
    capacity = 'Capacity'
    power = 'Power'
    availability = 'Availability'
    percent = 'Percent'
    control = 'Control'
    mode = 'Mode'


class ScheduleTypeLimit(IDdEnergyBaseModel):
    """Specifies the data types and limits for values contained in schedules."""

    type: constr(regex='^ScheduleTypeLimit$') = 'ScheduleTypeLimit'

    lower_limit: Union[NoLimit, float] = Field(
        default=NoLimit(),
        description='Lower limit for the schedule type or NoLimit.'
    )

    upper_limit: Union[NoLimit, float] = Field(
        default=NoLimit(),
        description='Upper limit for the schedule type or NoLimit.'
    )

    numeric_type: ScheduleNumericType = ScheduleNumericType.continuous

    unit_type: ScheduleUnitType = ScheduleUnitType.dimensionless

    class Config:
        @staticmethod
        def schema_extra(schema, model):
            schema['properties']['lower_limit']['anyOf'] = [
                    {"$ref": "#/components/schemas/NoLimit"},
                    {"type": "number"}
                ]
            schema['properties']['upper_limit']['anyOf'] = [
                    {"$ref": "#/components/schemas/NoLimit"},
                    {"type": "number"}
                ]


class ScheduleDay(IDdEnergyBaseModel):
    """Used to describe the daily schedule for a single simulation day."""

    type: constr(regex='^ScheduleDay$') = 'ScheduleDay'

    values: List[float] = Field(
        ...,
        description='A list of floats or integers for the values of the schedule. '
            'The length of this list must match the length of the times list.'
    )

    times: List[conlist(int, min_items=2, max_items=2)] = Field(
        [0, 0],
        description='A list of lists with each sub-list possesing 2 values for '
            '[hour, minute]. The length of the master list must match the length '
            'of the values list. Each time in the master list represents the time '
            'of day that the corresponding value begins to take effect. For example '
            '[(0,0), (9,0), (17,0)] in combination with the values [0, 1, 0] denotes a '
            'schedule value of 0 from 0:00 to 9:00, a value of 1 from 9:00 to 17:00 '
            'and 0 from 17:00 to the end of the day. Note that this representation '
            'of times as the "time of beginning" is a different convention than '
            'EnergyPlus, which uses "time until".'
    )

    @root_validator
    def check_times_values_match(cls, values):
        "Ensure the length of the times matches the length of the values."
        times, vals = values.get('times'), values.get('values')
        assert len(times) == len(vals), 'Length of schedule times must match ' \
            'the schedule values. {} != {}.'.format(len(v), len(values['values']))
        return values

    interpolate: bool = Field(
        False,
        description='Boolean to note whether values in between times should be '
            'linearly interpolated or whether successive values should take effect '
            'immediately upon the beginning time corrsponding to them.'
    )


class ScheduleRuleAbridged(DatedBaseModel):
    """Schedule rule including a ScheduleDay and when it should be applied.."""

    type: constr(regex='^ScheduleRuleAbridged$') = 'ScheduleRuleAbridged'

    schedule_day: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='The identifier of a ScheduleDay object associated with this rule.'
    )

    apply_sunday: bool = Field(
        False,
        description='Boolean noting whether to apply schedule_day on Sundays.'
    )

    apply_monday: bool = Field(
        False,
        description='Boolean noting whether to apply schedule_day on Mondays.'
    )

    apply_tuesday: bool = Field(
        False,
        description='Boolean noting whether to apply schedule_day on Tuesdays.'
    )

    apply_wednesday: bool = Field(
        False,
        description='Boolean noting whether to apply schedule_day on Wednesdays.'
    )

    apply_thursday: bool = Field(
        False,
        description='Boolean noting whether to apply schedule_day on Thursdays.'
    )

    apply_friday: bool = Field(
        False,
        description='Boolean noting whether to apply schedule_day on Fridays.'
    )

    apply_saturday: bool = Field(
        False,
        description='Boolean noting whether to apply schedule_day on Saturdays.'
    )

    start_date: List[int] = Field(
        [1, 1],
        min_items=2,
        max_items=3,
        description='A list of two integers for [month, day], representing the start '
            'date of the period over which the schedule_day will be applied.'
            'A third integer may be added to denote whether the date should be '
            're-serialized for a leap year (it should be a 1 in this case).'
    )

    @validator('start_date')
    def check_start_date(cls, v):
        return cls.check_date(v)

    end_date: List[int] = Field(
        [12, 31],
        min_items=2,
        max_items=3,
        description='A list of two integers for [month, day], representing the end '
            'date of the period over which the schedule_day will be applied.'
            'A third integer may be added to denote whether the date should be '
            're-serialized for a leap year (it should be a 1 in this case).'
    )

    @validator('end_date')
    def check_end_date(cls, v):
        return cls.check_date(v)


class ScheduleRulesetAbridged(IDdEnergyBaseModel):
    """Used to define a schedule for a default day, further described by ScheduleRule."""

    type: constr(regex='^ScheduleRulesetAbridged$') = 'ScheduleRulesetAbridged'

    day_schedules: List[ScheduleDay] = Field(
        ...,
        description='A list of ScheduleDays that are referenced in the other keys of '
            'this ScheduleRulesetAbridged.'
    )

    default_day_schedule: str = Field(
        ...,
        min_length=1,
        max_length=100,
        description='An identifier for the ScheduleDay that will be used for '
            'all days when no ScheduleRule is applied. This ScheduleDay must be '
            'in the day_schedules.'
    )

    schedule_rules: List[ScheduleRuleAbridged] = Field(
        default=None,
        description='A list of ScheduleRuleAbridged that note exceptions to the '
            'default_day_schedule. These rules should be ordered from highest to '
            'lowest priority.'
    )

    holiday_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='An identifier for the ScheduleDay that will be used for holidays. '
            'This ScheduleDay must be in the day_schedules.'
    )

    summer_designday_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='An identifier for the ScheduleDay that will be used for '
            'the summer design day. This ScheduleDay must be in the day_schedules.'
    )

    winter_designday_schedule: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='An identifier for the ScheduleDay that will be used for the '
            'winter design day. This ScheduleDay must be in the day_schedules.'
    )

    schedule_type_limit: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a ScheduleTypeLimit that will be used to validate '
            'schedule values against upper/lower limits and assign units to the '
            'schedule values. If None, no validation will occur.'
    )


class ScheduleRuleset(ScheduleRulesetAbridged):
    """Used to define a schedule for a default day, further described by ScheduleRule."""

    type: constr(regex='^ScheduleRuleset$') = 'ScheduleRuleset'

    schedule_type_limit: ScheduleTypeLimit = Field(
        default=None,
        description='ScheduleTypeLimit object that will be used to validate '
            'schedule values against upper/lower limits and assign units to the '
            'schedule values. If None, no validation will occur.',
    )


class ScheduleFixedIntervalAbridged(IDdEnergyBaseModel):
    """Used to specify a start date and a list of values for a period of analysis."""

    type: constr(regex='^ScheduleFixedIntervalAbridged$') = \
        'ScheduleFixedIntervalAbridged'

    values: List[float] = Field(
        ...,
        min_items=24,
        max_items=527040,
        description='A list of timeseries values occuring at each timestep over '
            'the course of the simulation.'
    )

    schedule_type_limit: str = Field(
        default=None,
        min_length=1,
        max_length=100,
        description='Identifier of a ScheduleTypeLimit that will be used to validate '
            'schedule values against upper/lower limits and assign units to the '
            'schedule values. If None, no validation will occur.'
    )

    timestep: int = Field(
        1,
        description='An integer for the number of steps per hour that the input '
            'values correspond to.  For example, if each value represents 30 '
            'minutes, the timestep is 2. For 15 minutes, it is 4.'
    )

    @validator('timestep')
    def check_timestep(cls, v):
        "Ensure the timestep is acceptable by EnergyPlus."
        valid_timesteps = (1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60)
        assert v in valid_timesteps, '"{}" is not a valid timestep. ' \
            'Choose from {}'.format(v, valid_timesteps)
        return v

    start_date: List[int] = Field(
        [1, 1],
        min_items=2,
        max_items=3,
        description='A list of two integers for [month, day], representing the start '
            'date when the schedule values begin to take effect.'
            'A third integer may be added to denote whether the date should be '
            're-serialized for a leap year (it should be a 1 in this case).'
    )

    @validator('start_date')
    def check_start_date(cls, v):
        if len(v) == 3 and v[2]:
            try:
                datetime.date(2016, v[0] , v[1])
            except ValueError:
                raise ValueError('{}/{} is not a valid date.'.format(v[0], v[1]))
        else:
            try:
                datetime.date(2017, v[0] , v[1])
            except ValueError:
                raise ValueError('{}/{} is not a valid date.'.format(v[0], v[1]))
        return v

    placeholder_value: float = Field(
        0,
        description=' A value that will be used for all times not covered by the '
            'input values. Typically, your simulation should not need to use this '
            'value if the input values completely cover the simulation period.'
    )

    interpolate: bool = Field(
        False,
        description='Boolean to note whether values in between intervals should be '
            'linearly interpolated or whether successive values should take effect '
            'immediately upon the beginning time corrsponding to them.'
    )

    @root_validator
    def check_number_of_values(cls, values):
        "Ensure an acceptable number of schedule values."
        vals = values.get('values')
        timestep = values.get('timestep')
        start_date = values.get('start_date')
        if len(vals) < 24 * timestep:
             raise ValueError('Number of schedule values must be for at least one day, '
                              'with a length greater than 24 * timestep.')
        max_l = timestep * 8760 if len(start_date) == 3 and start_date[2] \
            else timestep * 8784
        if len(vals) > max_l:
             raise ValueError('Number of schedule values must not exceed a full year, '
                              'with a length greater than 8760 * timestep.')
        if len(vals) % (24 * timestep) != 0:
            raise ValueError(
                'Number of schedule values must be for a whole number of days.')
        return values


class ScheduleFixedInterval(ScheduleFixedIntervalAbridged):
    """Used to specify a start date and a list of values for a period of analysis."""

    type: constr(regex='^ScheduleFixedInterval$') = 'ScheduleFixedInterval'

    schedule_type_limit: ScheduleTypeLimit = Field(
        default=None,
        description='ScheduleTypeLimit object that will be used to validate '
            'schedule values against upper/lower limits and assign units to the '
            'schedule values. If None, no validation will occur.',
    )


if __name__ == '__main__':
    print(ScheduleFixedIntervalAbridged.schema_json(indent=2))
    print(ScheduleRulesetAbridged.schema_json(indent=2))
    print(ScheduleFixedInterval.schema_json(indent=2))
    print(ScheduleRuleset.schema_json(indent=2))



# coding=utf-8
from __future__ import division

from honeybee_radiance.modifier.material import Trans, Light, BSDF

from honeybee_radiance.lib.modifiers import generic_wall, generic_ceiling, \
    black, generic_exterior_window, air_boundary, white_glow

import os
import json


def modifier_plastic_generic_wall(directory):
    dest_file = os.path.join(directory, 'modifier_plastic_generic_wall.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_wall.to_dict(), fp, indent=4)


def modifier_plastic_generic_ceiling(directory):
    dest_file = os.path.join(directory, 'modifier_plastic_generic_ceiling.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_ceiling.to_dict(), fp, indent=4)


def modifier_plastic_black(directory):
    dest_file = os.path.join(directory, 'modifier_plastic_black.json')
    with open(dest_file, 'w') as fp:
        json.dump(black.to_dict(), fp, indent=4)


def modifier_glass_generic_exterior_window(directory):
    dest_file = os.path.join(directory, 'modifier_glass_generic_exterior_window.json')
    with open(dest_file, 'w') as fp:
        json.dump(generic_exterior_window.to_dict(), fp, indent=4)


def modifier_glass_air_boundary(directory):
    dest_file = os.path.join(directory, 'modifier_glass_air_boundary.json')
    with open(dest_file, 'w') as fp:
        json.dump(air_boundary.to_dict(), fp, indent=4)


def modifier_trans_tree_foliage(directory):
    tree_leaves = Trans.from_single_reflectance('Foliage_0.3', 0.3, 0.0, 0.1, 0.15, 0.15)
    dest_file = os.path.join(directory, 'modifier_trans_tree_foliage.json')
    with open(dest_file, 'w') as fp:
        json.dump(tree_leaves.to_dict(), fp, indent=4)


def modifier_glow_white(directory):
    dest_file = os.path.join(directory, 'modifier_glow_white.json')
    with open(dest_file, 'w') as fp:
        json.dump(white_glow.to_dict(), fp, indent=4)


def modifier_light_green_spotlight(directory):
    green_spot = Light('Green_Spotlight', 0, 1, 0)
    dest_file = os.path.join(directory, 'modifier_light_green_spotlight.json')
    with open(dest_file, 'w') as fp:
        json.dump(green_spot.to_dict(), fp, indent=4)


def modifier_bsdf_klemsfull(directory):
    relative_path = './scripts/bsdf/klemsfull.xml'
    klemsfull = BSDF(relative_path)
    dest_file = os.path.join(directory, 'modifier_bsdf_klemsfull.json')
    json.dumps(klemsfull.to_dict(), indent=4)
    with open(dest_file, 'w') as fp:
        json.dump(klemsfull.to_dict(), fp, indent=4)


# run all functions within the file
master_dir = os.path.split(os.path.dirname(__file__))[0]
sample_directory = os.path.join(master_dir, 'samples', 'modifier')

modifier_plastic_generic_wall(sample_directory)
modifier_plastic_generic_ceiling(sample_directory)
modifier_plastic_black(sample_directory)
modifier_glass_generic_exterior_window(sample_directory)
modifier_glass_air_boundary(sample_directory)
modifier_trans_tree_foliage(sample_directory)
modifier_glow_white(sample_directory)
modifier_light_green_spotlight(sample_directory)
#modifier_bsdf_klemsfull(sample_directory)
